/*
 * Copyright 2012 uaiHebert Solucoes em Informatica
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * */
package com.uaihebert.model;


import com.uaihebert.util.ReflectionUtil;

import javax.persistence.EntityManager;
import javax.persistence.TypedQuery;
import javax.persistence.criteria.*;
import java.math.BigDecimal;
import java.util.*;
import java.util.logging.Logger;

@SuppressWarnings("unchecked")
public class BasicQueryElements<T> {

    private static final String ECLIPSELINK_BUG_MESSAGE_386354 = " \n There is a bug opened in the EclipseLink Jira. Give a vote to the bug issue, so it will be resolved faster: https://bugs.eclipse.org/bugs/show_bug.cgi?id=386354 ";
    private static final String HIBERNATE_BUG_MESSAGE_1022938 = " \n There is a bug opened in the Hibernate Jira. Give a vote to the bug issue, so it will be resolved faster: https://hibernate.onjira.com/browse/HHH-7985 ";
    private static final String OPENJPA_BUG_MESSAGE_2333 = " \n There is a bug opened in the OpenJPA Jira. Give a vote to the bug issue, so it will be resolved faster: https://issues.apache.org/jira/browse/OPENJPA-2333";

    private final Root<T> mainRoot;
    private final Class<T> entityClass;
    private final EntityManager entityManager;
    private final CriteriaQuery<T> criteriaQuery;
    private final CriteriaBuilder criteriaBuilder;
    private final List<Order> orders = new ArrayList<Order>();

    private final PredicatesManager predicatesManager = new PredicatesManager();

    private static final Logger LOG = Logger.getLogger("EasyCriteria");
    private final Map<String, Object> joins = new HashMap<String, Object>();

    public BasicQueryElements(CriteriaBuilder criteriaBuilder, CriteriaQuery<T> criteriaQuery, Root<T> mainRoot, EntityManager entityManager, Class<T> entityClass) {
        super();
        this.criteriaBuilder = criteriaBuilder;
        this.criteriaQuery = criteriaQuery;
        this.mainRoot = mainRoot;
        this.entityManager = entityManager;
        this.entityClass = entityClass;
    }

    public TypedQuery<T> getTypedQuery() {
        setUpParametersIfNeeded();
        setUpOrderByIfNeeded();

        return entityManager.createQuery(criteriaQuery);
    }

    private void setUpParametersIfNeeded() {
        if (predicatesManager.hasValues()) {
            criteriaQuery.where(predicatesManager.getAllPredicates(criteriaBuilder));
        }
    }

    private void setUpOrderByIfNeeded() {
        if (orders.isEmpty()) {
            return;
        }

        criteriaQuery.orderBy(orders);
    }

    private void isJoinCreated(String joinName) {
        if (joins.isEmpty()) {
            throw new IllegalStateException("Did you add any kind of join? You should call easyCriteria.join() method before adding any kind of join parameter.");
        }

        if (!joins.containsKey(joinName)) {
            throw new IllegalStateException("Ops, we could not find your Join. You should call easyCriteria.join() method before adding any kind of join parameter.");
        }
    }

    private boolean isOpenJPA() {
        return entityManager.toString().contains("openjpa");
    }

    private boolean isHibernate() {
        return entityManager.toString().contains("hibernate");
    }

    private boolean isEclipseLink() {
        return entityManager.toString().contains("eclipse");
    }

    @SuppressWarnings("rawtypes")
    private Path getJoinField(String joinName, String attributeName) {
        isJoinCreated(joinName);

        if (isOpenJPA()) {
            return mainRoot.get(joinName).get(attributeName);
        }

        Join join = (Join) joins.get(joinName);

        return join.get(attributeName);
    }

    private void addAndPredicate(Predicate andPredicate) {
        predicatesManager.addAndPredicate(andPredicate);
    }

    private void addOrPredicate(Predicate predicate) {
        predicatesManager.addOrPredicate(predicate);
    }

    private void addOrPredicate(List<Predicate> predicates) {
        addOrPredicate(1, predicates);
    }

    private void addOrPredicate(int index, List<Predicate> predicates) {
        for (Predicate predicate : predicates) {
            predicatesManager.addOrPredicate(index, predicate);
        }
    }

    private void addAndSeparatedByOr(int index, Predicate predicate) {
        if (isHibernate()) {
            String errorMessage = "This method is not implemented correctly on Hibernate yet (use Long instead of long). =/" + HIBERNATE_BUG_MESSAGE_1022938;
            LOG.severe(errorMessage);
        }

        predicatesManager.addAndSeparatedByOr(index, predicate);
    }

    public void andEquals(String attributeName, Object value) {
        try {
            addAndPredicate(criteriaBuilder.equal(mainRoot.get(attributeName), value));
        } catch (IllegalArgumentException e) {
            throw new IllegalArgumentException("We could not find the parameter: " + attributeName + " in the given class: " + entityClass.getCanonicalName());
        }
    }

    public void addAndSeparatedByOr(int index, String attributeName, Object value) {
        addAndSeparatedByOr(index, getEqualPredicateConverted(attributeName, value));
    }

    public void orEquals(String attributeName, Object... values) {
        if (isHibernate()) {
            String errorMessage = "This method is not implemented correctly on Hibernate yet (use Long instead of long). =/" + HIBERNATE_BUG_MESSAGE_1022938;
            LOG.severe(errorMessage);
        }

        addOrPredicate(extractOrPredicates(attributeName, values));
    }

    public void orEquals(int index, String attributeName, Object... values) {
        if (isHibernate()) {
            String errorMessage = "This method is not implemented correctly on Hibernate yet (use Long instead of long). =/" + HIBERNATE_BUG_MESSAGE_1022938;
            LOG.severe(errorMessage);
        }

        addOrPredicate(index, extractOrPredicates(attributeName, values));
    }

    private List<Predicate> extractOrPredicates(String attributeName, Object[] values) {
        List<Predicate> orPredicates = new ArrayList<Predicate>();

        for (Object value : values) {
            orPredicates.add(getEqualPredicateConverted(attributeName, value));
        }

        return orPredicates;
    }

    private Predicate getEqualPredicateConverted(String attributeName, Object value) {
        if (ReflectionUtil.isDouble(entityClass, attributeName)) {
            Double valueAsDouble = Double.parseDouble(value.toString());
            return criteriaBuilder.equal(mainRoot.get(attributeName).as(Double.class), valueAsDouble);
        }

        if (ReflectionUtil.isFloat(entityClass, attributeName)) {
            Float valueAsFloat = Float.parseFloat(value.toString());
            return criteriaBuilder.equal(mainRoot.get(attributeName).as(Float.class), valueAsFloat);
        }

        if (ReflectionUtil.isLong(entityClass, attributeName)) {
            Long valueAsLong = Long.parseLong(value.toString());
            return criteriaBuilder.equal(mainRoot.get(attributeName).as(Long.class), valueAsLong);
        }

        if (ReflectionUtil.isInteger(entityClass, attributeName)) {
            Integer valueAsInteger = Integer.parseInt(value.toString());
            return criteriaBuilder.equal(mainRoot.get(attributeName).as(Integer.class), valueAsInteger);
        }

        if (ReflectionUtil.isString(entityClass, attributeName)) {
            String valueAsString = value.toString();
            return criteriaBuilder.equal(mainRoot.get(attributeName).as(String.class), valueAsString);
        }

        if (ReflectionUtil.isDate(entityClass, attributeName)) {
            Date valueAsDate = (Date) value;
            return criteriaBuilder.equal(mainRoot.get(attributeName).as(Date.class), valueAsDate);
        }

        if (ReflectionUtil.isCalendar(entityClass, attributeName)) {
            Calendar valueAsCalendar = (Calendar) value;
            return criteriaBuilder.equal(mainRoot.get(attributeName).as(Calendar.class), valueAsCalendar);
        }

        if (ReflectionUtil.isBigDecimal(entityClass, attributeName)) {
            BigDecimal valueAsBigInteger = BigDecimal.valueOf(Double.parseDouble(value.toString()));
            return criteriaBuilder.equal(mainRoot.get(attributeName).as(BigDecimal.class), valueAsBigInteger);
        }

        return criteriaBuilder.equal(mainRoot.get(attributeName), value);
    }

    private Predicate[] extractNotPredicates(String attributeName, Object[] values) {
        Predicate[] orPredicates = new Predicate[values.length];

        for (int i = 0; i < values.length; i++) {
            orPredicates[i] = criteriaBuilder.not(criteriaBuilder.equal(mainRoot.get(attributeName), values[i]));
        }
        return orPredicates;
    }

    public void andNotEquals(String attributeName, Object value) {
        addAndPredicate(criteriaBuilder.and(criteriaBuilder.not(criteriaBuilder.equal(mainRoot.get(attributeName), value))));
    }

    public void orNotEquals(String attributeName, Object... values) {
        Predicate[] orPredicates = extractNotPredicates(attributeName, values);

        addOrPredicate(criteriaBuilder.or(orPredicates));
    }

    public void andGreaterThan(String attributeName, Object value) {
        if (ReflectionUtil.isDouble(entityClass, attributeName)) {
            Double valueAsDouble = Double.parseDouble(value.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.gt(mainRoot.get(attributeName).as(Double.class), valueAsDouble)));
            return;
        }

        if (ReflectionUtil.isFloat(entityClass, attributeName)) {
            Float valueAsFloat = Float.parseFloat(value.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.gt(mainRoot.get(attributeName).as(Float.class), valueAsFloat)));
            return;
        }

        if (ReflectionUtil.isLong(entityClass, attributeName)) {
            Long valueAsLong = (Long) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.gt(mainRoot.get(attributeName).as(Long.class), valueAsLong)));
            return;
        }

        if (ReflectionUtil.isInteger(entityClass, attributeName)) {
            Integer valueAsInteger = (Integer) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.gt(mainRoot.get(attributeName).as(Integer.class), valueAsInteger)));
            return;
        }

        if (ReflectionUtil.isString(entityClass, attributeName)) {
            String valueAsString = value.toString();
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.greaterThan(mainRoot.get(attributeName).as(String.class), valueAsString)));
            return;
        }

        if (ReflectionUtil.isDate(entityClass, attributeName)) {
            Date valueAsDate = (Date) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.greaterThan(mainRoot.get(attributeName).as(Date.class), valueAsDate)));
            return;
        }

        if (ReflectionUtil.isCalendar(entityClass, attributeName)) {
            Calendar valueAsCalendar = (Calendar) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.greaterThan(mainRoot.get(attributeName).as(Calendar.class), valueAsCalendar)));
            return;
        }

        if (ReflectionUtil.isBigDecimal(entityClass, attributeName)) {
            BigDecimal valueAsCalendar = (BigDecimal) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.gt(mainRoot.get(attributeName).as(BigDecimal.class), valueAsCalendar)));
            return;
        }

        throw new IllegalArgumentException("The attributeName: " + attributeName + " was not in the allowed types (int, double, float, long, calendar, date, big decimal). We looked inside the class: " + entityClass.getCanonicalName());
    }

    public void andGreaterOrEqualTo(String attributeName, Object value) {
        if (ReflectionUtil.isDouble(entityClass, attributeName)) {
            Double valueAsDouble = Double.parseDouble(value.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.ge(mainRoot.get(attributeName).as(Double.class), valueAsDouble)));
            return;
        }

        if (ReflectionUtil.isFloat(entityClass, attributeName)) {
            Float valueAsFloat = Float.parseFloat(value.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.ge(mainRoot.get(attributeName).as(Float.class), valueAsFloat)));
            return;
        }

        if (ReflectionUtil.isLong(entityClass, attributeName)) {
            Long valueAsLong = (Long) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.ge(mainRoot.get(attributeName).as(Long.class), valueAsLong)));
            return;
        }

        if (ReflectionUtil.isInteger(entityClass, attributeName)) {
            Integer valueAsLong = (Integer) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.ge(mainRoot.get(attributeName).as(Integer.class), valueAsLong)));
            return;
        }

        if (ReflectionUtil.isString(entityClass, attributeName)) {
            String valueAsString = value.toString();
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.greaterThanOrEqualTo(mainRoot.get(attributeName).as(String.class), valueAsString)));
            return;
        }

        if (ReflectionUtil.isDate(entityClass, attributeName)) {
            Date valueAsDate = (Date) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.greaterThanOrEqualTo(mainRoot.get(attributeName).as(Date.class), valueAsDate)));
            return;
        }

        if (ReflectionUtil.isCalendar(entityClass, attributeName)) {
            Calendar valueAsCalendar = (Calendar) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.greaterThanOrEqualTo(mainRoot.get(attributeName).as(Calendar.class), valueAsCalendar)));
            return;
        }

        if (ReflectionUtil.isBigDecimal(entityClass, attributeName)) {
            BigDecimal valueAsCalendar = (BigDecimal) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.ge(mainRoot.get(attributeName).as(BigDecimal.class), valueAsCalendar)));
            return;
        }

        throw new IllegalArgumentException("The attributeName: " + attributeName + " was not in the allowed types (int, double, float, long, calendar, date, big decimal). We looked inside the class: " + entityClass.getCanonicalName());
    }

    public void andLessThan(String attributeName, Object value) {
        if (ReflectionUtil.isDouble(entityClass, attributeName)) {
            Double valueAsDouble = Double.parseDouble(value.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lt(mainRoot.get(attributeName).as(Double.class), valueAsDouble)));
            return;
        }

        if (ReflectionUtil.isFloat(entityClass, attributeName)) {
            Float valueAsFloat = Float.parseFloat(value.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lt(mainRoot.get(attributeName).as(Float.class), valueAsFloat)));
            return;
        }

        if (ReflectionUtil.isLong(entityClass, attributeName)) {
            Long valueAsLong = (Long) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lt(mainRoot.get(attributeName).as(Long.class), valueAsLong)));
            return;
        }

        if (ReflectionUtil.isInteger(entityClass, attributeName)) {
            Integer valueAsLong = (Integer) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lt(mainRoot.get(attributeName).as(Integer.class), valueAsLong)));
            return;
        }

        if (ReflectionUtil.isString(entityClass, attributeName)) {
            String valueAsString = value.toString();
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lessThan(mainRoot.get(attributeName).as(String.class), valueAsString)));
            return;
        }

        if (ReflectionUtil.isDate(entityClass, attributeName)) {
            Date valueAsDate = (Date) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lessThan(mainRoot.get(attributeName).as(Date.class), valueAsDate)));
            return;
        }

        if (ReflectionUtil.isCalendar(entityClass, attributeName)) {
            Calendar valueAsCalendar = (Calendar) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lessThan(mainRoot.get(attributeName).as(Calendar.class), valueAsCalendar)));
            return;
        }

        if (ReflectionUtil.isBigDecimal(entityClass, attributeName)) {
            BigDecimal valueAsBigDecimal = (BigDecimal) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lt(mainRoot.get(attributeName).as(BigDecimal.class), valueAsBigDecimal)));
            return;
        }

        throw new IllegalArgumentException("The attributeName: " + attributeName + " was not in the allowed types (int, double, float, long, calendar, date, big decimal). We looked inside the class: " + entityClass.getCanonicalName());
    }

    public void andLessOrEqualTo(String attributeName, Object value) {
        if (ReflectionUtil.isDouble(entityClass, attributeName)) {
            Double valueAsDouble = Double.parseDouble(value.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.le(mainRoot.get(attributeName).as(Double.class), valueAsDouble)));
            return;
        }

        if (ReflectionUtil.isFloat(entityClass, attributeName)) {
            Float valueAsFloat = Float.parseFloat(value.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.le(mainRoot.get(attributeName).as(Float.class), valueAsFloat)));
            return;
        }

        if (ReflectionUtil.isLong(entityClass, attributeName)) {
            Long valueAsLong = (Long) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.le(mainRoot.get(attributeName).as(Long.class), valueAsLong)));
            return;
        }

        if (ReflectionUtil.isInteger(entityClass, attributeName)) {
            Integer valueAsLong = (Integer) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.le(mainRoot.get(attributeName).as(Integer.class), valueAsLong)));
            return;
        }

        if (ReflectionUtil.isString(entityClass, attributeName)) {
            String valueAsString = value.toString();
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lessThanOrEqualTo(mainRoot.get(attributeName).as(String.class), valueAsString)));
            return;
        }

        if (ReflectionUtil.isDate(entityClass, attributeName)) {
            Date valueAsDate = (Date) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lessThanOrEqualTo(mainRoot.get(attributeName).as(Date.class), valueAsDate)));
            return;
        }

        if (ReflectionUtil.isCalendar(entityClass, attributeName)) {
            Calendar valueAsCalendar = (Calendar) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lessThanOrEqualTo(mainRoot.get(attributeName).as(Calendar.class), valueAsCalendar)));
            return;
        }

        if (ReflectionUtil.isBigDecimal(entityClass, attributeName)) {
            BigDecimal valueAsCalendar = (BigDecimal) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.le(mainRoot.get(attributeName).as(BigDecimal.class), valueAsCalendar)));
            return;
        }

        throw new IllegalArgumentException("The attributeName: " + attributeName + " was not in the allowed types (int, double, float, long, calendar, date, big decimal). We looked inside the class: " + entityClass.getCanonicalName());
    }

    public void setDistinctTrue() {
        criteriaQuery.distinct(true);
    }

    public <J> void innerJoin(String joinName) {
        if (isOpenJPA()) {
            String errorMessage = "This method is not implemented correctly on OpenJPA yet. =/" + OPENJPA_BUG_MESSAGE_2333;
            LOG.severe(errorMessage);
        }

        try {
            Join<T, J> join = mainRoot.join(joinName, JoinType.INNER);
            joins.put(joinName, join);
        } catch (IllegalArgumentException e) {
            throw new IllegalArgumentException("We could not find the join: " + joinName + " in the given class: " + entityClass.getCanonicalName());
        }
    }

    public <J> void leftJoin(String joinName) {
        if (isOpenJPA()) {
            String errorMessage = "This method is not implemented correctly on OpenJPA yet. =/" + OPENJPA_BUG_MESSAGE_2333;
            LOG.severe(errorMessage);
        }

        try {
            Join<T, J> join = mainRoot.join(joinName, JoinType.LEFT);
            joins.put(joinName, join);
        } catch (IllegalArgumentException e) {
            throw new IllegalArgumentException("We could not find the join: " + joinName + " in the given class: " + entityClass.getCanonicalName());
        }
    }

    public <J> void innerJoinFetch(String joinName) {
        try {
            Fetch<T, J> fetch = mainRoot.fetch(joinName, JoinType.INNER);
            joins.put(joinName, fetch);
        } catch (IllegalArgumentException e) {
            throw new IllegalArgumentException("We could not find the join: " + joinName + " in the given class: " + entityClass.getCanonicalName());
        }
    }

    public <J> void leftJoinFetch(String joinName) {
        try {
            Fetch<T, J> fetch = mainRoot.fetch(joinName, JoinType.LEFT);
            joins.put(joinName, fetch);
        } catch (IllegalArgumentException e) {
            throw new IllegalArgumentException("We could not find the join: " + joinName + " in the given class: " + entityClass.getCanonicalName());
        }
    }

    public void andBetween(String attributeName, Object valueA, Object valueB) {
        if (ReflectionUtil.isDouble(entityClass, attributeName)) {
            Double convertedValueA = Double.parseDouble(valueA.toString());
            Double convertedValueB = Double.parseDouble(valueB.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.between(mainRoot.get(attributeName).as(Double.class), convertedValueA, convertedValueB)));
            return;
        }

        if (ReflectionUtil.isFloat(entityClass, attributeName)) {
            Float convertedValueA = Float.parseFloat(valueA.toString());
            Float convertedValueB = Float.parseFloat(valueB.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.between(mainRoot.get(attributeName).as(Float.class), convertedValueA, convertedValueB)));
            return;
        }

        if (ReflectionUtil.isLong(entityClass, attributeName)) {
            Long convertedValueA = (Long) valueA;
            Long convertedValueB = (Long) valueB;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.between(mainRoot.get(attributeName).as(Long.class), convertedValueA, convertedValueB)));
            return;
        }

        if (ReflectionUtil.isInteger(entityClass, attributeName)) {
            Integer convertedValueA = (Integer) valueA;
            Integer convertedValueB = (Integer) valueB;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.between(mainRoot.get(attributeName).as(Integer.class), convertedValueA, convertedValueB)));
            return;
        }

        if (ReflectionUtil.isString(entityClass, attributeName)) {
            String convertedValueA = valueA.toString();
            String convertedValueB = valueB.toString();
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.between(mainRoot.get(attributeName).as(String.class), convertedValueA, convertedValueB)));
            return;
        }

        if (ReflectionUtil.isDate(entityClass, attributeName)) {
            Date convertedValueA = (Date) valueA;
            Date convertedValueB = (Date) valueB;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.between(mainRoot.get(attributeName).as(Date.class), convertedValueA, convertedValueB)));
            return;
        }

        if (ReflectionUtil.isCalendar(entityClass, attributeName)) {
            Calendar convertedValueA = (Calendar) valueA;
            Calendar convertedValueB = (Calendar) valueB;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.between(mainRoot.get(attributeName).as(Calendar.class), convertedValueA, convertedValueB)));
            return;
        }

        if (ReflectionUtil.isBigDecimal(entityClass, attributeName)) {
            BigDecimal convertedValueA = (BigDecimal) valueA;
            BigDecimal convertedValueB = (BigDecimal) valueB;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.between(mainRoot.get(attributeName).as(BigDecimal.class), convertedValueA, convertedValueB)));
            return;
        }

        throw new IllegalArgumentException("The attributeName: " + attributeName + " was not in the allowed types (int, double, float, long, calendar, date, big decimal). We looked inside the class: " + entityClass.getCanonicalName());
    }

    public void andIsNull(String attributeName) {
        addAndPredicate(criteriaBuilder.and(criteriaBuilder.isNull(mainRoot.get(attributeName))));
    }

    public void andIsNotNull(String attributeName) {
        addAndPredicate(criteriaBuilder.and(criteriaBuilder.isNotNull(mainRoot.get(attributeName))));
    }

    public void andCollectionIsEmpty(String collectionName) {
        if (ReflectionUtil.isList(entityClass, collectionName)) {
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.isEmpty(mainRoot.get(collectionName).as(List.class))));
            return;
        }

        if (ReflectionUtil.isSet(entityClass, collectionName)) {
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.isEmpty(mainRoot.get(collectionName).as(Set.class))));
            return;
        }

        if (ReflectionUtil.isCollection(entityClass, collectionName)) {
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.isEmpty(mainRoot.get(collectionName).as(Collection.class))));
            return;
        }

        throw new IllegalArgumentException("The collection: " + collectionName + " was not in the allowed types (list, collection or set). We looked inside the class: " + entityClass.getCanonicalName());
    }


    public void andCollectionIsNotEmpty(String collectionName) {
        if (ReflectionUtil.isList(entityClass, collectionName)) {
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.isNotEmpty(mainRoot.get(collectionName).as(List.class))));
            return;
        }

        if (ReflectionUtil.isSet(entityClass, collectionName)) {
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.isNotEmpty(mainRoot.get(collectionName).as(Set.class))));
            return;
        }

        if (ReflectionUtil.isCollection(entityClass, collectionName)) {
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.isNotEmpty(mainRoot.get(collectionName).as(Collection.class))));
            return;
        }

        throw new IllegalArgumentException("The collection: " + collectionName + " was not in the allowed types (list, collection or set). We looked inside the class: " + entityClass.getCanonicalName());
    }

    public void whereStringLike(String attributeName, String value) {
        addAndPredicate(criteriaBuilder.and(criteriaBuilder.like(mainRoot.get(attributeName).as(String.class), value)));
    }

    public void whereStringNotLike(String attributeName, String value) {
        addAndPredicate(criteriaBuilder.and(criteriaBuilder.notLike(mainRoot.get(attributeName).as(String.class), value)));
    }

    public void andStringIn(String attributeName, List<String> values) {
        addAndPredicate(criteriaBuilder.and(mainRoot.get(attributeName).as(String.class).in(values)));
    }

    public void andStringNotIn(String attributeName, List<String> values) {
        addAndPredicate(criteriaBuilder.and(criteriaBuilder.not(mainRoot.get(attributeName).as(String.class).in(values))));
    }

    public void orderByAsc(String attributeName) {
        orders.add(criteriaBuilder.asc(mainRoot.get(attributeName)));
    }

    public void orderByDesc(String attributeName) {
        orders.add(criteriaBuilder.desc(mainRoot.get(attributeName)));
    }

    public void andJoinEquals(String joinName, String attributeName, Object value) {
        addAndPredicate(criteriaBuilder.and(criteriaBuilder.equal(getJoinField(joinName, attributeName), value)));
    }

    public void andJoinNotEquals(String joinName, String attributeName, Object value) {
        addAndPredicate(criteriaBuilder.and(criteriaBuilder.not(criteriaBuilder.equal(getJoinField(joinName, attributeName), value))));
    }

    private <J> String getJoinCanonicalName(String joinName) {
        isJoinCreated(joinName);

        if (isOpenJPA()) {
            Path<Object> objectPath = mainRoot.get(joinName);

            if (ReflectionUtil.isAcceptableCollection(objectPath.getJavaType())) {
                return objectPath.getModel().getBindableJavaType().getCanonicalName();
            }

            return objectPath.getJavaType().getCanonicalName();
        }

        Join<T, J> join = (Join<T, J>) joins.get(joinName);

        return join.getJavaType().getCanonicalName();
    }

    private Path<Integer> getIntegerJoinField(String joinName, String attributeName) {
        isJoinCreated(joinName);

        if (isOpenJPA()) {
            return mainRoot.get(joinName).get(attributeName);
        }

        Join<T, Integer> join = (Join<T, Integer>) joins.get(joinName);

        return join.get(attributeName);
    }

    private Path<Long> getLongJoinField(String joinName, String attributeName) {
        isJoinCreated(joinName);

        if (isOpenJPA()) {
            return mainRoot.get(joinName).get(attributeName);
        }

        Join<T, Long> join = (Join<T, Long>) joins.get(joinName);

        return join.get(attributeName);
    }

    private Path<Double> getDoubleJoinField(String joinName, String attributeName) {
        isJoinCreated(joinName);

        if (isOpenJPA()) {
            return mainRoot.get(joinName).get(attributeName);
        }

        Join<T, Double> join = (Join<T, Double>) joins.get(joinName);

        return join.get(attributeName);
    }

    private Path<Float> getFloatJoinField(String joinName, String attributeName) {
        isJoinCreated(joinName);

        if (isOpenJPA()) {
            return mainRoot.get(joinName).get(attributeName);
        }

        Join<T, Float> join = (Join<T, Float>) joins.get(joinName);

        return join.get(attributeName);
    }

    private Path<Date> getDateJoinField(String joinName, String attributeName) {
        isJoinCreated(joinName);

        if (isOpenJPA()) {
            return mainRoot.get(joinName).get(attributeName);
        }

        Join<T, Date> join = (Join<T, Date>) joins.get(joinName);

        return join.get(attributeName);
    }

    private Path<Calendar> getCalendarJoinField(String joinName, String attributeName) {
        isJoinCreated(joinName);

        if (isOpenJPA()) {
            return mainRoot.get(joinName).get(attributeName);
        }

        Join<T, Calendar> join = (Join<T, Calendar>) joins.get(joinName);

        return join.get(attributeName);
    }

    private Path<String> getStringJoinField(String joinName, String attributeName) {
        isJoinCreated(joinName);

        if (isOpenJPA()) {
            return mainRoot.get(joinName).get(attributeName);
        }

        Join<T, String> join = (Join<T, String>) joins.get(joinName);

        return join.get(attributeName);
    }

    private Path<BigDecimal> getBigDecimalJoinField(String joinName, String attributeName) {
        isJoinCreated(joinName);

        if (isOpenJPA()) {
            return mainRoot.get(joinName).get(attributeName);
        }

        Join<T, BigDecimal> join = (Join<T, BigDecimal>) joins.get(joinName);

        return join.get(attributeName);
    }

    public void andJoinGreaterThan(String joinName, String attributeName, Object value) {
        if (ReflectionUtil.isJoinAttributeInteger(getJoinCanonicalName(joinName), attributeName)) {
            Integer valueConverted = (Integer) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.gt(getIntegerJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeLong(getJoinCanonicalName(joinName), attributeName)) {
            Long valueConverted = (Long) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.gt(getLongJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeDouble(getJoinCanonicalName(joinName), attributeName)) {
            Double valueConverted = Double.parseDouble(value.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.gt(getDoubleJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeFloat(getJoinCanonicalName(joinName), attributeName)) {
            Float valueConverted = Float.parseFloat(value.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.gt(getFloatJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeDate(getJoinCanonicalName(joinName), attributeName)) {
            Date valueConverted = (Date) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.greaterThan(getDateJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeCalendar(getJoinCanonicalName(joinName), attributeName)) {
            Calendar valueConverted = (Calendar) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.greaterThan(getCalendarJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeString(getJoinCanonicalName(joinName), attributeName)) {
            String valueConverted = value.toString();
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.greaterThan(getStringJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeBigDecimal(getJoinCanonicalName(joinName), attributeName)) {
            BigDecimal valueConverted = (BigDecimal) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.gt(getBigDecimalJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        throw new IllegalArgumentException("The attributeName: " + attributeName + " was not in the allowed types (int, double, float, long, calendar, date, big decimal). We looked inside the given class: " + joinName);
    }

    public void andJoinGreaterOrEqualTo(String joinName, String attributeName, Object value) {
        if (ReflectionUtil.isJoinAttributeInteger(getJoinCanonicalName(joinName), attributeName)) {
            Integer valueConverted = Integer.valueOf(value.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.ge(getIntegerJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeLong(getJoinCanonicalName(joinName), attributeName)) {
            Long valueConverted = (Long) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.ge(getLongJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeDouble(getJoinCanonicalName(joinName), attributeName)) {
            Double valueConverted = Double.parseDouble(value.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.ge(getDoubleJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeFloat(getJoinCanonicalName(joinName), attributeName)) {
            Float valueConverted = Float.parseFloat(value.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.ge(getFloatJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeDate(getJoinCanonicalName(joinName), attributeName)) {
            Date valueConverted = (Date) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.greaterThanOrEqualTo(getDateJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeCalendar(getJoinCanonicalName(joinName), attributeName)) {
            Calendar valueConverted = (Calendar) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.greaterThanOrEqualTo(getCalendarJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeString(getJoinCanonicalName(joinName), attributeName)) {
            String valueConverted = value.toString();
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.greaterThanOrEqualTo(getStringJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeBigDecimal(getJoinCanonicalName(joinName), attributeName)) {
            BigDecimal valueConverted = (BigDecimal) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.ge(getBigDecimalJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        throw new IllegalArgumentException("The attributeName: " + attributeName + " was not in the allowed types (int, double, float, long, calendar, date, big decimal). We looked inside the given class: " + joinName);
    }

    public void andJoinLessThan(String joinName, String attributeName, Object value) {
        if (ReflectionUtil.isJoinAttributeInteger(getJoinCanonicalName(joinName), attributeName)) {
            Integer valueConverted = Integer.valueOf(value.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lt(getIntegerJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeLong(getJoinCanonicalName(joinName), attributeName)) {
            Long valueConverted = (Long) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lt(getLongJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeDouble(getJoinCanonicalName(joinName), attributeName)) {
            Double valueConverted = Double.parseDouble(value.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lt(getDoubleJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeFloat(getJoinCanonicalName(joinName), attributeName)) {
            Float valueConverted = Float.parseFloat(value.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lt(getFloatJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeDate(getJoinCanonicalName(joinName), attributeName)) {
            Date valueConverted = (Date) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lessThan(getDateJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeCalendar(getJoinCanonicalName(joinName), attributeName)) {
            Calendar valueConverted = (Calendar) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lessThan(getCalendarJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeString(getJoinCanonicalName(joinName), attributeName)) {
            String valueConverted = value.toString();
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lessThan(getStringJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeBigDecimal(getJoinCanonicalName(joinName), attributeName)) {
            BigDecimal valueConverted = (BigDecimal) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lt(getBigDecimalJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        throw new IllegalArgumentException("The attributeName: " + attributeName + " was not in the allowed types (int, double, float, long, calendar, date, big decimal). We looked inside the given class: " + joinName);
    }

    public void andJoinLessOrEqualTo(String joinName, String attributeName, Object value) {
        if (ReflectionUtil.isJoinAttributeInteger(getJoinCanonicalName(joinName), attributeName)) {
            Integer valueConverted = Integer.valueOf(value.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.le(getIntegerJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeLong(getJoinCanonicalName(joinName), attributeName)) {
            Long valueConverted = (Long) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.le(getLongJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeDouble(getJoinCanonicalName(joinName), attributeName)) {
            Double valueConverted = Double.parseDouble(value.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.le(getDoubleJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeFloat(getJoinCanonicalName(joinName), attributeName)) {
            Float valueConverted = Float.parseFloat(value.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.le(getFloatJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeDate(getJoinCanonicalName(joinName), attributeName)) {
            Date valueConverted = (Date) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lessThanOrEqualTo(getDateJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeCalendar(getJoinCanonicalName(joinName), attributeName)) {
            Calendar valueConverted = (Calendar) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lessThanOrEqualTo(getCalendarJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeString(getJoinCanonicalName(joinName), attributeName)) {
            String valueConverted = value.toString();
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.lessThanOrEqualTo(getStringJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeBigDecimal(getJoinCanonicalName(joinName), attributeName)) {
            BigDecimal valueConverted = (BigDecimal) value;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.le(getBigDecimalJoinField(joinName, attributeName), valueConverted)));
            return;
        }

        throw new IllegalArgumentException("The attributeName: " + attributeName + " was not in the allowed types (int, double, float, long, calendar, date, big decimal). We looked inside the given class: " + joinName);
    }

    public void andJoinBetween(String joinName, String attributeName, Object valueA, Object valueB) {
        if (ReflectionUtil.isJoinAttributeDouble(getJoinCanonicalName(joinName), attributeName)) {
            Double convertedValueA = Double.parseDouble(valueA.toString());
            Double convertedValueB = Double.parseDouble(valueB.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.between(getJoinField(joinName, attributeName).as(Double.class), convertedValueA, convertedValueB)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeFloat(getJoinCanonicalName(joinName), attributeName)) {
            Float convertedValueA = Float.parseFloat(valueA.toString());
            Float convertedValueB = Float.parseFloat(valueB.toString());
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.between(getJoinField(joinName, attributeName).as(Float.class), convertedValueA, convertedValueB)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeLong(getJoinCanonicalName(joinName), attributeName)) {
            Long convertedValueA = (Long) valueA;
            Long convertedValueB = (Long) valueB;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.between(getJoinField(joinName, attributeName).as(Long.class), convertedValueA, convertedValueB)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeInteger(getJoinCanonicalName(joinName), attributeName)) {
            Integer convertedValueA = (Integer) valueA;
            Integer convertedValueB = (Integer) valueB;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.between(getJoinField(joinName, attributeName).as(Integer.class), convertedValueA, convertedValueB)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeString(getJoinCanonicalName(joinName), attributeName)) {
            String convertedValueA = valueA.toString();
            String convertedValueB = valueB.toString();
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.between(getJoinField(joinName, attributeName).as(String.class), convertedValueA, convertedValueB)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeDate(getJoinCanonicalName(joinName), attributeName)) {
            Date convertedValueA = (Date) valueA;
            Date convertedValueB = (Date) valueB;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.between(getJoinField(joinName, attributeName).as(Date.class), convertedValueA, convertedValueB)));
            return;
        }

        if (ReflectionUtil.isJoinAttributeCalendar(getJoinCanonicalName(joinName), attributeName)) {
            Calendar convertedValueA = (Calendar) valueA;
            Calendar convertedValueB = (Calendar) valueB;
            addAndPredicate(criteriaBuilder.and(criteriaBuilder.between(getJoinField(joinName, attributeName).as(Calendar.class), convertedValueA, convertedValueB)));
            return;
        }

        throw new IllegalArgumentException("The attributeName: " + attributeName + " was not in the allowed types (int, double, float, long, calendar, date, big decimal). We looked inside the given class: " + joinName);
    }

    public void andJoinAttributeIsNull(String joinName, String attributeName) {
        addAndPredicate(criteriaBuilder.and(criteriaBuilder.isNull(getJoinField(joinName, attributeName))));
    }

    public void andJoinAttributeIsNotNull(String joinName, String attributeName) {
        addAndPredicate(criteriaBuilder.and(criteriaBuilder.isNotNull(getJoinField(joinName, attributeName))));
    }

    public void andJoinListIsEmpty(String joinName, String listName) {
        if (isEclipseLink()) {
            String errorMessage = "This method is not implemented correctly on EclipseLink yet. =/" + ECLIPSELINK_BUG_MESSAGE_386354;
            LOG.severe(errorMessage);
        }

        addAndPredicate(criteriaBuilder.and(criteriaBuilder.isEmpty(getJoinField(joinName, listName).as(List.class))));
    }

    public void andJoinCollectionIsEmpty(String joinName, String collectionName) {
        if (isEclipseLink()) {
            String errorMessage = "This method is not implemented correctly on EclipseLink, because of a bug of this implementation. =/" + ECLIPSELINK_BUG_MESSAGE_386354;
            LOG.severe(errorMessage);
        }

        addAndPredicate(criteriaBuilder.and(criteriaBuilder.isEmpty(getJoinField(joinName, collectionName).as(Collection.class))));
    }

    public void andJoinSetIsEmpty(String joinName, String setName) {
        if (isEclipseLink()) {
            String errorMessage = "This method is not implemented correctly on EclipseLink yet. =/" + ECLIPSELINK_BUG_MESSAGE_386354;
            LOG.severe(errorMessage);
        }

        addAndPredicate(criteriaBuilder.and(criteriaBuilder.isEmpty(getJoinField(joinName, setName).as(Set.class))));
    }

    public void andJoinListIsNotEmpty(String joinName, String listName) {
        addAndPredicate(criteriaBuilder.and(criteriaBuilder.isNotEmpty(getJoinField(joinName, listName).as(List.class))));
    }

    public void andJoinCollectionIsNotEmpty(String joinName, String collectionName) {
        addAndPredicate(criteriaBuilder.and(criteriaBuilder.isNotEmpty(getJoinField(joinName, collectionName).as(Collection.class))));
    }

    public void andJoinSetIsNotEmpty(String joinName, String setName) {
        addAndPredicate(criteriaBuilder.and(criteriaBuilder.isNotEmpty(getJoinField(joinName, setName).as(Set.class))));
    }

    public void andJoinStringLike(String joinName, String attributeName, String value) {
        addAndPredicate(criteriaBuilder.and(criteriaBuilder.like((getJoinField(joinName, attributeName).as(String.class)), value)));
    }


    public void andJoinStringNotLike(String joinName, String attributeName, String value) {
        addAndPredicate(criteriaBuilder.and(criteriaBuilder.notLike((getJoinField(joinName, attributeName).as(String.class)), value)));
    }

    public void andJoinStringIn(String joinName, String attributeName, List<String> values) {
        addAndPredicate(criteriaBuilder.and((getJoinField(joinName, attributeName).as(String.class).in(values))));
    }

    public void andJoinStringNotIn(String joinName, String attributeName, List<String> values) {
        addAndPredicate(criteriaBuilder.and(criteriaBuilder.not(getJoinField(joinName, attributeName).as(String.class).in(values))));
    }
}