/*
 * Copyright 2013 uaiHebert Solucoes em Informatica
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * */
package com.uaihebert.model;

import java.util.List;

/**
 * @author uaiHebert.com
 *
 * <T> Class to be queried. Can be used like EasyCriteria<Dog> easyCriteria
 *
 */
public interface EasyCriteria<T> {

    /**
     * Will use the EntityManager.getResultList() method to return the data.
     *
     * @return a list of objects
     */
    public List<T> getResultList();

    /**
     * Will use the EntityManager.getSingleResult() method to return the data. <br/> <br/>
     *
     * <b> The JPA might raise the NonUniqueResultException or the NoResultException</b>
     *
     * @return just one object
     */
    public T getSingleResult();

    /**
     * Method that uses the "=" of the JPQL. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p where p.name = 'Joseph'</code> <br/><br/>
     *
     * This method can be used with any kind of "simple" attribute.<br/>
     * <b>This method should not be used with
     * relationships (OneToOne, OneToMany, ManyToOne, ManyToMany), ElementCollection, etc.</b>
     *
     * @param attributeName the class attribute name
     * @param value to be used in the query
     * @return the current EasyCriteria instance
     */
    public EasyCriteriaImp<T> andEquals(String attributeName, Object value);

    /**
     * Method that uses the "OR" of the JPQL. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p where p.name = 'Joseph' OR p.name = 'Mary' OR p.name = 'Mark'</code> <br/><br/>
     *
     * This method can be used with any kind of "simple" attribute.<br/>
     * <b>This method should not be used with
     * relationships (OneToOne, OneToMany, ManyToOne, ManyToMany), ElementCollection, etc.</b>
     *
     * @param attributeName the class attribute name
     * @param values to be used in the query
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> orEquals(String attributeName, Object... values);

    /**
     * Method that uses "OR" combined with "AND" expression of the JPQL. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p<br/>
     * where p.name = 'Joseph' OR p.name = 'Mary'<br/>
     * and p.age = 33 or p.age = 43</code> <br/><br/>
     *
     * To create the query above would be created with the methods: <br/>
     * <code>
     *      easyCriteria.orEquals(1, "name", "Joseph").orEquals(1, "name", "Mary").orEquals(2, "age", "33").orEquals(2, "age", "43");
     * </code><br/><br/>
     * For each group of "ors" an index is used. <br/><br/>
     *
     * If your implementation is Hibernate, you should not use attribute long, use Long instead. There is a open bug, please vote for it: https://hibernate.onjira.com/browse/HHH-7985  <br/><br/>
     *
     * This method can be used with any kind of "simple" attribute.<br/>
     * <b>This method should not be used with
     * relationships (OneToOne, OneToMany, ManyToOne, ManyToMany), ElementCollection, etc.</b>
     *
     * @param index the grouped or order
     * @param attributeName the class attribute name
     * @param values to be used in the query
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> orEquals(int index, String attributeName, Object... values);

    /**
     * Method that uses the "<>" of the JPQL. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p where p.name <> 'Joseph'</code> <br/><br/>
     *
     * This method can be used with any kind of "simple" attribute.<br/>
     * <b>This method should not be used with
     * relationships (OneToOne, OneToMany, ManyToOne, ManyToMany), ElementCollection, etc.</b>
     *
     * @param attributeName the class attribute name
     * @param value to be used in the query
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andNotEquals(String attributeName, Object value);

    /**
     * Method that uses the "<>" with OR of the JPQL. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p where p.name <> 'Anna' or p.name <> 'Mary'</code> <br/><br/>
     *
     * This method can be used with any kind of "simple" attribute.<br/>
     * <b>This method should not be used with
     * relationships (OneToOne, OneToMany, ManyToOne, ManyToMany), ElementCollection, etc.</b>
     *
     * @param attributeName the class attribute name
     * @param values to be used in the query
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> orNotEquals(String attributeName, Object... values);

    /**
     * Method that uses the ">" of the JPQL with Double attributes. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p where p.weight > 10.4d</code> <br/><br/>
     *
     * You can use this method with a double, float, integer, long, date, calendar, string<br/><br/>
     *
     * This method can be used with any kind of "simple" attribute.<br/>
     * <b>This method should not be used with
     * relationships (OneToOne, OneToMany, ManyToOne, ManyToMany), ElementCollection, etc.</b>
     *
     * @param attributeName the class attribute name
     * @param value to be used in the query
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andGreaterThan(String attributeName, Object value);

    /**
     * Method that uses the ">=" of the JPQL with Double attributes. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p where p.weight >= 10.4d</code> <br/><br/>
     *
     * You can use this method with a double, float, integer, long, date, calendar, string<br/><br/>
     *
     * This method can be used with any kind of "simple" attribute.<br/>
     * <b>This method should not be used with
     * relationships (OneToOne, OneToMany, ManyToOne, ManyToMany), ElementCollection, etc.</b>
     *
     * @param attributeName the class attribute name
     * @param value to be used in the query
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andGreaterOrEqualTo(String attributeName, Object value);

    /**
     * Method that uses the "<" of the JPQL with Double attributes. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p where p.weight < 10.4d</code> <br/><br/>
     *
     * You can use this method with a double, float, integer, long, date, calendar, string<br/><br/>
     *
     * This method can be used with any kind of "simple" attribute.<br/>
     * <b>This method should not be used with
     * relationships (OneToOne, OneToMany, ManyToOne, ManyToMany), ElementCollection, etc.</b>
     *
     * @param attributeName the class attribute name
     * @param value to be used in the query
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andLessThan(String attributeName, Object value);

    /**
     * Method that uses the "<=" of the JPQL with Double attributes. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p where p.weight < 10.4d</code> <br/><br/>
     *
     * You can use this method with a double, float, integer, long, date, calendar, string<br/><br/>
     *
     * This method can be used with any kind of "simple" attribute.<br/>
     * <b>This method should not be used with
     * relationships (OneToOne, OneToMany, ManyToOne, ManyToMany), ElementCollection, etc.</b>
     *
     * @param attributeName the class attribute name
     * @param value to be used in the query
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andLessOrEqualTo(String attributeName, Object value);


    /**
     * Will do a inner join with a class relationship. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p inner join p.dogs d</code> <br/><br/>
     *
     * <b>If your implementation is OpenJPA, use the setDistinctTrue(), vote this bug: https://issues.apache.org/jira/browse/OPENJPA-2333</b>
     *
     * @param joinName the relationship to be joined
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> innerJoin(String joinName);

    /**
     * Will do a left join with a class relationship. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p left join p.dogs d</code> <br/><br/>
     *
     * <b>If your implementation is OpenJPA, use the setDistinctTrue(), vote this bug: https://issues.apache.org/jira/browse/OPENJPA-2333</b>
     *
     * @param joinName the relationship to be joined
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> leftJoin(String joinName);

    /**
     * Will do a inner join fetch with a class relationship.A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p inner join fetch p.dogs d</code> <br/><br/>
     *
     * @param joinName the relationship to be joined
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> innerJoinFetch(String joinName);

    /**
     * Will do a inner join fetch with a class relationship.A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p inner join fetch p.dogs d</code> <br/><br/>
     *
     * @param joinName the relationship to be joined
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> leftJoinFetch(String joinName);

    /**
     * Will use the distinct word in the query. A JPQL like bellow might be created: <br/><br/>
     * <code>select distinct p from Person p</code> <br/><br/>
     *
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> setDistinctTrue();


    /**
     * Method that uses the "between" of the JPQL with object attributes. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p where p.age between 19 and 29</code> <br/><br/>
     *
     * You can use this method with a double, float, integer, long, date, calendar, string<br/><br/>
     *
     * This method can be used with any kind of "simple" attribute.<br/>
     * <b>This method should not be used with
     * relationships (OneToOne, OneToMany, ManyToOne, ManyToMany), ElementCollection, etc.</b>
     *
     * @param attributeName the class attribute name
     * @param valueA the first number
     * @param valueB the last number
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andBetween(String attributeName, Object valueA, Object valueB);

    /**
     * Will check if a field is null.<br/><br/> A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p where p.name is null</code> <br/><br/>
     *
     * <b> The JPA will raise exception if this method is used with a Collection </b>
     *
     * @param attributeName the class attribute name
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andIsNull(String attributeName);

    /**
     * Will check if a field is not null.<br/><br/> A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p where p.name is not null</code> <br/><br/>
     *
     * <b> The JPA will raise exception if this method is used with a Collection </b>
     *
     * @param attributeName the class attribute name
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andIsNotNull(String attributeName);

    /**
     * Will check if a Collection is empty.<br/><br/>A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p where p.dogs is empty</code> <br/><br/>
     *
     * This method should be used with java.util.Collection, java.util.Set and java.util.List <br/>
     *
     * <b> The JPA may raise exception if this method is used with a non collection attribute or bring unexpected results </b>
     *
     * @param collectionName the class Collection name
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andCollectionIsEmpty(String collectionName);

    /**
     * Will check if a Collection is not empty.<br/><br/>A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p where p.dogs is not empty</code> <br/><br/>
     *
     * This method should be used with java.util.Collection, java.util.Set and java.util.List <br/>
     *
     * <b> The JPA may raise exception if this method is used with a non collection attribute or bring unexpected results </b>
     *
     * @param collectionName the class Collection name
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andCollectionIsNotEmpty(String collectionName);

    /**
     * Will do a String Like. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p where p.name like 'Mar%'</code> <br/><br/>
     *
     * <b> You must add the % symbol like:</b>
     * <ul>
     *              <li>Mar%</li>
     *              <li>%ar</li>
     *              <li>%ar%</li>
     * </ul>
     *
     * @param attributeName the class attribute name
     * @param value value to be used as parameter
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andStringLike(String attributeName, String value);

    /**
     * Will do a String Not Like. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p where p.name not like 'Mar%'</code> <br/><br/>
     *
     * <b> You must add the % symbol like:</b>
     * <ul>
     *              <li>Mar%</li>
     *              <li>%ar</li>
     *              <li>%ar%</li>
     * </ul>
     *
     * @param attributeName the class attribute name
     * @param value value to be used as parameter
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andStringNotLike(String attributeName, String value);

    /**
     * Will do a String in. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p where p.name in ('A', 'B')</code> <br/><br/>
     *
     * @param attributeName the class attribute name
     * @param values value to be used as parameter
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andStringIn(String attributeName, List<String> values);

    /**
     * Will do a String not in. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p where p.name not in ('A', 'B')</code> <br/><br/>
     *
     * @param attributeName the class attribute name
     * @param values value to be used as parameter
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andStringNotIn(String attributeName, List<String> values);

    /**
     * Will order your query result. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p order by p.name asc</code> <br/><br/>
     *
     * @param attributeName the attribute name
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> orderByAsc(String attributeName);

    /**
     * Will order your query result. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p order by p.name desc</code> <br/><br/>
     *
     * @param attributeName the attribute name
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> orderByDesc(String attributeName);

    /**
     * Set the first Result of the query result
     * @param firstResult the first result index
     */
    public void setFirstResult(int firstResult);

    /**
     * Set the max result to be returned
     * @param maxResults the max results to be returned
     */
    public void setMaxResults(int maxResults);

    /**
     * Method that uses the "=" of the JPQL in a join clausule. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p join p.dogs d where d.name = 'Dark'</code> <br/><br/>
     *
     * This method can be used with any kind of "simple" attribute.
     * <b>This method should not be used with
     * relationships (OneToOne, OneToMany, ManyToOne, ManyToMany), ElementCollection, etc.</b>
     *
     * @param joinName the join name to receive the where values
     * @param attributeName the class attribute name
     * @param value to be used in the query
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andJoinEquals(String joinName, String attributeName, Object value);

    /**
     * Method that uses the "<>" of the JPQL in a join clausule. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p join p.dogs d where d.name <> 'Dark'</code> <br/><br/>
     *
     * This method can be used with any kind of "simple" attribute.<br/>
     * <b>This method should not be used with
     * relationships (OneToOne, OneToMany, ManyToOne, ManyToMany), ElementCollection, etc.</b>
     *
     * @param joinName the join name to receive the where values
     * @param attributeName the class attribute name
     * @param value to be used in the query
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andJoinNotEquals(String joinName, String attributeName, Object value);

    /**
     * Method that uses the ">" of the JPQL with Integer attributes. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p join p.dogs d where d.weight > 10</code> <br/><br/>
     *
     * You can use this method with a double, float, integer, long, date, calendar, string<br/><br/>
     *
     * This method can be used with any kind of "simple" attribute.<br/>
     * <b>This method should not be used with
     * relationships (OneToOne, OneToMany, ManyToOne, ManyToMany), ElementCollection, etc.</b>
     *
     * @param joinName the joined relationship to receive the parameter
     * @param attributeName the class attribute name
     * @param value to be used in the query
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andJoinGreaterThan(String joinName, String attributeName, Object value);

    /**
     * Method that uses the ">=" of the JPQL with Integer attributes. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p join p.dogs d where d.weight >= 10</code> <br/><br/>
     *
     * You can use this method with a double, float, integer, long, date, calendar, string<br/><br/>
     *
     * This method can be used with any kind of "simple" attribute.<br/>
     * <b>This method should not be used with
     * relationships (OneToOne, OneToMany, ManyToOne, ManyToMany), ElementCollection, etc.</b>
     *
     * @param joinName the joined relationship to receive the parameter
     * @param attributeName the class attribute name
     * @param value to be used in the query
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andJoinGreaterOrEqualTo(String joinName, String attributeName, Object value);

    /**
     * Method that uses the "<" of the JPQL with Integer attributes. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p join p.dogs d where d.weight < 10</code> <br/><br/>
     *
     * You can use this method with a double, float, integer, long, date, calendar, string<br/><br/>
     *
     * This method can be used with any kind of "simple" attribute.<br/>
     * <b>This method should not be used with
     * relationships (OneToOne, OneToMany, ManyToOne, ManyToMany), ElementCollection, etc.</b>
     *
     * @param joinName the joined relationship to receive the parameter
     * @param attributeName the class attribute name
     * @param value to be used in the query
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andJoinLessThan(String joinName, String attributeName, Object value);

    /**
     * Method that uses the "<=" of the JPQL with Integer attributes. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p join p.dogs d where d.weight <= 10</code> <br/><br/>
     *
     * You can use this method with a double, float, integer, long, date, calendar, string<br/><br/>
     *
     * This method can be used with any kind of "simple" attribute.<br/>
     * <b>This method should not be used with
     * relationships (OneToOne, OneToMany, ManyToOne, ManyToMany), ElementCollection, etc.</b>
     *
     * @param joinName the joined relationship to receive the parameter
     * @param attributeName the class attribute name
     * @param value to be used in the query
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andJoinLessOrEqualTo(String joinName, String attributeName, Object value);

    /**
     * Method that uses the "between" of the JPQL with Integer attributes. A JPQL like bellow might be created: <br/><br/>
     * <code>select d from Dog d join d.person p where p.age between 19 and 29</code> <br/><br/>
     *
     * You can use this method with a double, float, integer, long, date, calendar, string<br/><br/>
     *
     * This method can be used with any kind of "simple" attribute.<br/>
     * <b>This method should not be used with
     * relationships (OneToOne, OneToMany, ManyToOne, ManyToMany), ElementCollection, etc.</b>
     *
     * @param joinName the joined relationship to receive the parameter
     * @param attributeName the class attribute name
     * @param valueA the first number
     * @param valueB the last number
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andJoinBetween(String joinName, String attributeName, Object valueA, Object valueB);

    /**
     * Will check if a field is null.<br/><br/> A JPQL like bellow might be created: <br/><br/>
     * <code>select d from Dog d join d.person p where p.name is null</code> <br/><br/>
     *
     * <b> The JPA will raise exception if this method is used with a Collection </b>
     *
     * @param joinName the joined relationship to receive the parameter
     * @param attributeName the class attribute name
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andJoinAttributeIsNull(String joinName, String attributeName);

    /**
     * Will check if a field is not null.<br/><br/> A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p where p.name is not null</code> <br/><br/>
     *
     * <b> The JPA will raise exception if this method is used with a Collection </b>
     *
     * @param joinName the joined relationship to receive the parameter
     * @param attributeName the class attribute name
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andJoinAttributeIsNotNull(String joinName, String attributeName);

    /**
     * Will check  if a Collection is empty.<br/><br/>A JPQL like bellow might be created: <br/><br/>
     * <code>select c from Car c join c.person p where p.dogs is empty</code> <br/><br/>
     *
     * <b> The JPA may raise exception if this method is used with a non collection attribute or bring unexpected results </b>
     *
     * @param joinName the joined relationship to receive the parameter
     * @param listName the class Collection name
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andJoinListIsEmpty(String joinName, String listName);

    /**
     * @see #andJoinAttributeIsNotNull(String, String)
     *
     * */
    public EasyCriteria<T> andJoinCollectionIsEmpty(String joinName, String collectionName);

    /**
     * @see #andJoinAttributeIsNotNull(String, String)
     *
     * */
    public EasyCriteria<T> andJoinSetIsEmpty(String joinName, String setName);

    /**
     * Will check  if a collection is not empty.<br/><br/>A JPQL like bellow might be created: <br/><br/>
     * <code>select c from Car c join c.person p where p.dogs is not empty</code> <br/><br/>
     *
     * <b> The JPA may raise exception if this method is used with a non collection attribute or bring unexpected results </b>
     *
     * @param joinName the joined relationship to receive the parameter
     * @param listName Field to be checked
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andJoinListIsNotEmpty(String joinName, String listName);

    /**
     * @see #andJoinListIsNotEmpty(String, String)
     *
     * */
    public EasyCriteria<T> andJoinCollectionIsNotEmpty(String joinName, String collectionName);

    /**
     * @see #andJoinListIsNotEmpty(String, String)
     *
     * */
    public EasyCriteria<T> andJoinSetIsNotEmpty(String joinName, String setName);

    /**
     * Will do a String Like. A JPQL like bellow might be created: <br/><br/>
     * <code>select d from Dog d join d.person p where p.name like 'Mar%'</code> <br/><br/>
     *
     * <b> You must add the % symbol like:</b>
     * <ul>
     * 		<li>Mar%</li>
     * 		<li>%ar</li>
     * 		<li>%ar%</li>
     * </ul>
     *
     * @param joinName the joined relationship to receive the parameter
     * @param attributeName the class attribute name
     * @param value value to be used as parameter
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andJoinStringLike(String joinName, String attributeName, String value);

    /**
     * Will do a String Not Like. A JPQL like bellow might be created: <br/><br/>
     * <code>select d from Dog d join d.person p where p.name not like 'Mar%'</code> <br/><br/>
     *
     * <b> You must add the % symbol like:</b>
     * <ul>
     * 		<li>Mar%</li>
     * 		<li>%ar</li>
     * 		<li>%ar%</li>
     * </ul>
     *
     * @param joinName the joined relationship to receive the parameter
     * @param attributeName the class attribute name
     * @param value value to be used as parameter
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andJoinStringNotLike(String joinName, String attributeName, String value);

    /**
     * Will do a String in. A JPQL like bellow might be created: <br/><br/>
     * <code>select d from Dog d join d.person p where p.name in ('A', 'B')</code> <br/><br/>
     *
     * @param joinName the joined relationship to receive the parameter
     * @param attributeName the class attribute name
     * @param values to be used as parameter
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andJoinStringIn(String joinName, String attributeName, List<String> values);

    /**
     * Will do a String not in. A JPQL like bellow might be created: <br/><br/>
     * <code>select d from Dog d join d.person where p.name not in ('A', 'B')</code> <br/><br/>
     *
     * @param joinName the joined relationship to receive the parameter
     * @param attributeName the class attribute name
     * @param values to be used as parameter
     * @return the current EasyCriteria instance
     */
    public EasyCriteria<T> andJoinStringNotIn(String joinName, String attributeName, List<String> values);

    /**
     * Method that will use "AND" separated with "OR" expression of the JPQL. A JPQL like bellow might be created: <br/><br/>
     * <code>select p from Person p<br/>
     * where p.name = 'Joseph' and p.age = 33<br/>
     * or p.name = 'Mary' and p.age = 43</code> <br/><br/>
     *
     * The query above would be created with the methods: <br/>
     * <code>
     *      easyCriteria.addAndSeparatedByOr(1, "name", "Joseph").addAndSeparatedByOr(1, "age", "33").addAndSeparatedByOr(2, "name", "Mary").addAndSeparatedByOr(2, "age", "43");
     * </code><br/><br/>
     * For each group of "ors" an index is used. <br/><br/>
     *
     * If your implementation is Hibernate, you should not use attribute long, use Long instead. There is a open bug, please vote for it: https://hibernate.onjira.com/browse/HHH-7985 <br/><br/>
     *
     * This method can be used with any kind of "simple" attribute.<br/><br/>
     * <b>This method should not be used with
     * relationships (OneToOne, OneToMany, ManyToOne, ManyToMany), ElementCollection, etc.</b>
     *
     * @param index the grouped or order
     * @param attributeName the class attribute name
     * @param value to be used in the query
     * @return the current EasyCriteria instance
     */
   public EasyCriteria<T> addAndSeparatedByOr(int index, String attributeName, Object value);
}